# -*- coding: utf-8 -*-

"""
Created on Fri Jun 14 16:16:30 2019

@author: Luc RINAUDO
"""

import tkinter as tk
from tkinter import font as tkfont

import pydot
from PIL import Image,ImageTk


# INITIALISATION
def init_graph() :
    gr=pydot.Dot(graph_type = 'digraph', nodesep = 0.5, ranksep = 0.25)
    gr.set_node_defaults(shape = "circle", style = "filled", width = 0.5, height = 0.5)
    gr.set_edge_defaults(color = "blue", arrowhead = "vee")
    return gr

# ARBRE

""" Tree : [ etiquette , [ sous-arbre gauche, sous-arbre droit ] ] """

def string2forest(mess):
    """
    Entrée : String
    Sortie : [[n1, car1], [n2, car2] .....]
    Donne le tableau des effectifs des caractères d'une chaîne
    """
    fr = []
    while mess !="":
        car = mess[0]
        fr.append([mess.count(car),car])
        mess = mess.replace(car,"")
    return fr

def forest2foresttree(fr,opt=False):
    frtr = list(fr)
    while len(frtr) > 1 :
        frtr = huffman_step(frtr,opt)
    return frtr

def tree_height(tr, acc = 0):
    """
    Entrée : Tree
    Sorite : Int
    Donne la profondeur d'un arbre // recursion terminale
    """
    if isinstance(tr[1], str):
        return acc
    else:
        acc += 1
        return max(tree_height((tr[1][0]), acc), tree_height((tr[1][1]), acc))
        
def last_step(mess):
    n = 0
    while mess != "":
        car = mess[0]
        mess = mess.replace(car,"")
        n += 1
    return n-1

def huffman_step(old_fr,opt):
    fr = list(old_fr)
    if len(fr) > 1 :
        for tr in fr :
            tr.append(tree_height(tr))
        fr.sort(key = lambda x: (x[0], (-1)**int(not opt)*x[2]))        
        fr[0]=[fr[0][0]+fr[1][0], [fr[0][:2], fr[1][:2]], max([fr[0][2], fr[1][2]])+1]
        fr.pop(1)
        for tr in fr :
            tr.pop()
    return fr

# GRAPHE

def forest2graph(fr,dic=False,gr=""):
    if gr == "" :
        gr = init_graph()        
        
    for i in range(len(fr)):
        tree2graph(fr[i], dic, gr, str(i)+"r")
    return gr

def tree2graph(tr,dic=False,gr="",node_name='0r' ):   
    
    if gr == "" :
        gr = init_graph()        
        
    if isinstance(tr[1],str) :
        node_shape = "record"
        if dic :
            node_label = "{"+str(tr[0])+"| {"+tr[1]+" | "+node_name.split("r")[1]+"}}"
        else :
            node_label = "{"+str(tr[0])+"| "+tr[1]+"}"
        node_width = 0.37
        node_color = "green"
        
    else :
        node_shape = "circle"
        node_label = str(tr[0])
        node_width = .5
        node_color = "yellow"
    
    # Création du noeud
    node=pydot.Node(node_name,shape=node_shape,label=node_label,width=node_width,fillcolor=node_color)
    gr.add_node(node)
    
    # Creation arête   
    
    if node_name[-1] != "r" : # si noeud non racine
        parent_name = "\""+node_name[:-1]+"\""
        parent = gr.get_node(parent_name)[0]
        
        if dic :
            label_edge = node_name[-1]
        else :
            label_edge = " "
            
        edge = pydot.Edge(parent, node, label=label_edge)
        gr.add_edge(edge)
        
    # recursion
        
    if isinstance(tr[1], list) :
        for i in range(2) :                
            tree2graph(tr[1][i], dic, gr, node_name+str(i))
    return gr

# CODAGE

def huffman_code(tree, code="", dico={}):
    if isinstance(tree[1], str):
        dico[tree[1]] = code
    else:
        for i in range(2):
            huffman_code(tree[1][i], code+str(i), dico)
    return dico        

def huffman_decode(tree):
    decode = {}
    code = huffman_code(tree)
    for key in code :
        decode[code[key]] = key
    return decode

def codage(mess_str, opt=False):
    code=huffman_code(forest2foresttree(string2forest(mess_str), opt)[0])
    coded_str = ""
    for i in mess_str :
        coded_str += str(code[i])
    return coded_str

def decodage(coded_str, dico):
    decoded_str = ""
    i = 1
    while coded_str != "":
        while coded_str[:i] not in dico:
            i+=1
        decoded_str+=dico[coded_str[:i]]
        coded_str=coded_str[i:]
        i = 1 
    return decoded_str

# GUI

def update_message():
    last = last_step(message.get())
    scale.configure(to=last, command=update_step)
    step.set(0)
    
    text_codage.delete('1.0', 'end')
    text_codage.insert('1.0', codage(message.get(), opt_height.get()))

    update_graph()

def update_graph():
    h=step.get()
    update_step(h)
    
def update_step(h):
    opt = opt_height.get()
    dic = add_dict.get()
    mess = message.get()
    
    label_forest.config(text = "Étape "+str(h))
    
    text_codage.grid_forget()
    
    last = last_step(mess)
    if int(h) == last :
        update_final()
            
    forest = string2forest(mess)
    for s in range(int(h)):
        forest = huffman_step(forest,opt)
    graph = forest2graph(forest,dic)
    refresh(graph)
    
def update_final():
    label_forest.config(text="Arbre de Huffman")
    text_codage.grid(padx=10)
    
def refresh(gr):
    gr.write("huffman.png", prog="dot", format="png")
    gr.write("huffman"+str(step.get())+".png", prog="dot", format="png")
    photo = tk.PhotoImage(file="huffman.png")
    h = photo.height()
    w = photo.width()
    h = max(h,hc)
    w = max(w,wc)
    canvas.config(width=w, height=h)
    canvas.itemconfig(image, image=photo)
    canvas.coords(image, w//2, h//2)
    canvas.image = photo

mainwin = tk.Tk()

default_font = tkfont.nametofont("TkDefaultFont")
entry_font = tkfont.nametofont("TkTextFont")
default_font.configure(size=12)
entry_font.configure(size=16)

#mainwin.geometry("1024x768+100+100")

mainwin.title("Codage de Huffman")

label_mess = tk.Label(mainwin, text="Phrase à compresser")

message = tk.StringVar()
entry_mess = tk.Entry(mainwin,textvariable=message)

button_mess = tk.Button(mainwin, text="OK", command=update_message)

label_algo = tk.Label(mainwin, text="Algorithme")

wc = 640
hc = 400
canvas = tk.Canvas(mainwin, bg="white", width=wc, height=hc)
blank = ImageTk.PhotoImage(Image.new("RGB", (wc, hc), "white"))
image = canvas.create_image(wc//2, hc//2, image=blank)

step = tk.IntVar()
scale = tk.Scale(mainwin, variable=step, from_=0, to=0)

opt_height = tk.BooleanVar()
check_opt = tk.Checkbutton(mainwin, text="Opt. Hauteur", variable = opt_height, command = update_graph)

add_dict = tk.BooleanVar()
check_dic = tk.Checkbutton(mainwin, text="Dictionnaire", variable = add_dict, command = update_graph)

label_codage = tk.Label(mainwin, text="Phrase compressée")
text_codage = tk.Text(mainwin,font=("TkDefaultFont", 16),height=1,width=36)

label_forest = tk.Label(mainwin)

label_mess.grid(row = 0, column = 0, sticky=tk.W, padx=20, pady=20)

entry_mess.grid(row = 1,column = 0,columnspan = 2)
button_mess.grid(row = 1,column = 2)

label_algo.grid(row = 2, column = 0, sticky = tk.W, padx = 20, pady = 20)

canvas.grid(row = 3, column = 0, rowspan = 3,columnspan = 2, padx=10)

scale.grid(row = 3,column = 2)

check_opt.grid(row = 4, column = 2)
check_dic.grid(row = 5,column = 2)

label_forest.grid(row = 6, column = 0, columnspan = 2)

label_codage.grid(row = 7, column=0, sticky = tk.W, padx = 20, pady = 20)
text_codage.grid(row = 8, column = 0, columnspan = 2)
text_codage.grid_forget()

mainwin.mainloop()